require 'emcee'
require 'emcee/cache'

module ResourceManagement

  NO_NODESET_ERROR = "No nodeset was found corresponding to the submitted nodeset_id"

  KEY_TYPE           = 'type'
  KEY_TITLE          = 'title'
  KEY_PARAMS         = 'parameters'
  KEY_NODE           = 'node_name'
  KEY_AVOID_CONFLICT = 'avoid_conflict'

  TYPE_GROUP         = 'group'
  TYPE_USER          = 'user'

  PARAM_UID          = :uid
  PARAM_GID          = :gid
  PARAM_GROUP        = :group
  PARAM_GROUPS       = :groups

  class << self

    def retrieve_nodeset(nodeset_id)
      cache_strategy = Emcee::Cache.strategy
      Emcee::Cache.set_strategy(nil)
      cache = Emcee::Cache.new(nodeset_id)
      out = cache.load
      Emcee::Cache.set_strategy(cache_strategy)
      out
    end

    def generate_nodeset_id(nodes)
      nodeset_id = Digest::MD5.new.hexdigest(nodes.sort.join)
      cache_strategy = Emcee::Cache.strategy
      Emcee::Cache.set_strategy(nil)
      # setting expires_in to 0 results in no expiration
      # which is fine for nodesets since a nodes list should always
      # hash to the same nodeset_id
      cache = Emcee::Cache.new(nodeset_id, :expires_in => 0)
      cache.save(nodes.sort)
      Emcee::Cache.set_strategy(cache_strategy)
      nodeset_id
    end

    def string_to_nodes_regex(s)
      # Disallow regular expressions.
      chars = s.split('')
      if chars.first == '/' && chars.last == '/'
        s = s[1..-2]
      end
      # Allow shell-style globbing with * and ?.
      if s.include?('*') || s.include?('?')
        s = Regexp.escape(s)
        s.gsub!('\*', '.*')
        s.gsub!('\?', '.')
        s = '/^' + s + '$/'
      end
      return s
    end

    def resources_to_variations(raw_resources)
      resources_hash = Hash.new { |hash, key| hash[key] = Hash.new { |h,k| h[k] = [] } }
      raw_resources.each { |r| resources_hash[r[KEY_TITLE]][r[KEY_PARAMS]] << r[KEY_NODE] }
      variation_count = 0
      resources = resources_hash.map do |title, variations|
        variation_count += variations.size
        properties = []
        node_count = 0
        variations = variations.collect do |variation, nodes|
          properties |= variation.keys
          node_count += nodes.size
          { :attributes => variation, :nodes => nodes }
        end
        {
          'title'      => title,
          'variations' => variations.to_a,
          'nodeCount'  => node_count,
          'properties' => properties,
        }
      end
      [resources, raw_resources.length, variation_count]
    end

    def enforce_parameter_whitelist!(resource)
      type = resource[KEY_TYPE]
      parameters = resource[KEY_PARAMS]
      return if !type || !parameters
      case type
      when TYPE_USER
        # Remove non-system-agnostic user parameters.
        valid_params = [:name, :comment, PARAM_GROUP, PARAM_GROUPS, :password, :home, :shell, :ensure]
        parameters.each do |param, value|
          parameters.delete param unless valid_params.include? param
        end
      end
    end

    def fetch(query_nodes, type, title=nil, frontend_encode=true)
      mc_client = Emcee.new(:nodeset => query_nodes, :cache_strategy => :none)
      raw_resources = title ? mc_client.resource(type, title) : mc_client.resources(type)
      frontend_encode_resources!(query_nodes, raw_resources) if frontend_encode
      resources_to_variations(raw_resources)
    end

    def empty_clone_results()
      {:successes => [], :failures => [], :errors => []}
    end

    #
    # @param  target_nodes An array of node names to which the resource should be
    #                      cloned.
    # @param  resource A hash containing the keys 'type', 'title', and 'parameters'.
    # @return A hash containing the keys :successes, :failures, and :errors, each
    #         of which has an array of node names as a value. All elements of the
    #         nodes argument will be represented in this hash exactly once.
    #
    def clone(target_nodes, resource)
      return empty_clone_results if target_nodes.empty?

      # Work with a deep copy since the resource will be changed in place.
      resource = resource_deep_copy(resource)
      param_keys_to_symbols!(resource)

      case resource[KEY_TYPE]
      when TYPE_USER
        clone_user(target_nodes, resource)
      when TYPE_GROUP
        clone_group(target_nodes, resource)
      else
        shallow_clone(target_nodes, resource)
      end
    end

    def frontend_encode_resources!(nodes, raw_resources)
      raw_resources.each { |r| r[KEY_TYPE].downcase! if r.has_key? KEY_TYPE }
      # Users.
      users = raw_resources.select { |r| r[KEY_TYPE] == TYPE_USER }
      if !users.empty?
        mc_client = Emcee.new(:nodeset => nodes, :cache_strategy => :none)
        frontend_encode_users!(users, mc_client.resources(TYPE_GROUP))
      end
      # Groups.
      groups = raw_resources.select { |r| r[KEY_TYPE] == TYPE_GROUP }
      frontend_encode_groups!(groups) unless groups.empty?
    end

    private

    def clone_group(target_nodes, resource)
      resource[KEY_AVOID_CONFLICT] = PARAM_GID.to_s
      shallow_clone(target_nodes, resource)
    end

    def clone_user(target_nodes, resource)
      # Frontend decode user.
      user_params = resource[KEY_PARAMS]
      primary_group_name = user_params[PARAM_GROUP]
      secondary_group_names = user_params[PARAM_GROUPS] || []
      resource[KEY_AVOID_CONFLICT] = PARAM_UID.to_s
      resource[KEY_PARAMS].delete PARAM_GROUP

      secondary_group_names.each do |group_name|
        clone(target_nodes, bare_group_resource(group_name))
      end

      if primary_group_name && !primary_group_name.strip.empty?
        clone_user_with_primary_group(target_nodes, resource, primary_group_name.strip)
      else
        shallow_clone(target_nodes, resource)
      end
    end

    def clone_user_with_primary_group(target_nodes, resource, primary_group_name)
      primary_group = bare_group_resource(primary_group_name.strip)
      primary_group_results = shallow_clone(target_nodes, primary_group, false)
      results = empty_clone_results
      primary_group_results[:failures].each do |f|
        f[:reason] = "Could not create user's primary group. " + f[:reason]
        results[:failures] << f
      end
      primary_group_results[:errors].each do |e|
        e[:reason] = "Error creating user's primary group. " + e[:reason]
        results[:errors] << e
      end
      # Do a piece-meal clone of the user based on the results
      # of primary group clone and collate the results.
      primary_group_results[:successes].each do |success|
        success['variations'].each do |variant|
          resource[KEY_PARAMS][PARAM_GID] = variant[:attributes][PARAM_GID]
          partial_results = shallow_clone(variant[:nodes], resource)
          results[:successes] += partial_results[:successes]
          results[:failures] += partial_results[:failures]
          results[:errors] += partial_results[:errors]
        end
      end
      results[:successes] = consolidate_successes(results[:successes])
      return results
    end

    def consolidate_successes(successes)
      return [] if successes.empty?
      variations_hash = Hash.new { |hash, key| hash[key] = [] }
      successes.each do |s|
        s['variations'].each do |variation|
          variations_hash[variation[:attributes]] += variation[:nodes]
        end
      end

      result = successes.first.dup
      result['variations'] = variations_hash.map {|k,v| { :attributes => k, :nodes => v }}
      [ result ]
    end

    def bare_group_resource(group_name)
      {
        KEY_TYPE           => TYPE_GROUP,
        KEY_TITLE          => group_name,
        KEY_AVOID_CONFLICT => PARAM_GID.to_s,
        KEY_PARAMS         => { :ensure => :present }
      }
    end

    def frontend_encode_users!(users, groups)
      # Sort groups by gid and node name for fast lookup.
      groups_by_gid = groups.group_by { |g| g[KEY_PARAMS] ? g[KEY_PARAMS][PARAM_GID] : nil }
      groups_by_gid_and_node_name = {}
      groups_by_gid.each do |gid, groups|
        groups_by_gid_and_node_name[gid] = groups.group_by { |g| g[KEY_NODE] }
      end
      # Create a group parameter that has the group name corresponding
      # to the user's gid. Remove the uid and gid parameters.
      users.each do |user|
        parameters = user[KEY_PARAMS]
        next unless parameters
        parameters.delete PARAM_UID
        gid = parameters.delete PARAM_GID
        next unless gid
        gid_groups = groups_by_gid_and_node_name[gid]
        if gid_groups
          groups = gid_groups[user[KEY_NODE]]
          if groups && groups.length == 1
            parameters[PARAM_GROUP] = groups.first[KEY_TITLE]
            next
          end
        end
      end
    end

    def frontend_encode_groups!(groups)
      groups.each do |group|
        parameters = group[KEY_PARAMS]
        parameters.delete PARAM_GID if parameters
      end
    end

    def shallow_clone(target_nodes, resource, frontend_encode=true)
      # Work with a deep copy since keys are changed from strings to
      # symbols in place.
      resource = resource_deep_copy(resource)

      mc_client = Emcee.new(:nodeset => target_nodes, :cache_strategy => :none)
      results = empty_clone_results
      raw_results = mc_client.create_resource(resource)
      successes = []
      raw_results.each do |raw_result|
        result = {
          :nodeName  => raw_result.node_name,
        }
        if raw_result.status_code == 0
          if raw_result.raw_contents[:status] == "Resource was created"
            successful_clone = raw_result.raw_contents[:resource]
            successful_clone['node_name'] = raw_result.node_name
            successes << successful_clone
          else
            result[:reason] = raw_result.raw_contents[:status]
            results[:failures] << result
          end
        else
          result[:reason] = raw_result.status_message
          results[:errors] << result
        end
      end
      frontend_encode_resources!(target_nodes, successes) if frontend_encode
      results[:successes] = resources_to_variations(successes).first
      results
    end

    def resource_deep_copy(resource)
      out = resource.dup
      if out.has_key? KEY_PARAMS
        params = out[KEY_PARAMS].dup
        params.each do |key, value|
          params[key] = value.dup if value.class == Array
        end
        out[KEY_PARAMS] = params
      end
      out
    end

    def param_keys_to_symbols!(resource)
      if resource.has_key?(KEY_PARAMS)
        params = {}
        resource[KEY_PARAMS].each do |key, value|
          params[key.to_sym] = value
        end
        resource[KEY_PARAMS] = params
      end
    end

  end
end
